@@ -1,4 +1,5 @@
 # Copyright (C) 2013-2014 Science and Technology Facilities Council.
+# Copyright (C) 2015-2016 East Asian Observatory.
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,6 +23,7 @@ MAX_ORDER = 29
 
 MOC_TYPES = ('IMAGE', 'CATALOG')
 
+
 class MOC(object):
     """Class representing Multi-Order Coverage maps.
 
@@ -34,8 +36,8 @@ class MOC(object):
     """
 
     def __init__(self, order=None, cells=None,
-            filename=None, filetype=None,
-            name=None, mocid=None, origin=None, moctype=None):
+                 filename=None, filetype=None,
+                 name=None, mocid=None, origin=None, moctype=None):
         """Construct new MOC object.
 
         By default the new MOC will be empty, but if an order and a
@@ -72,16 +74,27 @@ class MOC(object):
         self._orders = tuple(set() for i in range(0, MAX_ORDER + 1))
         self._normalized = True
 
+        # Initialize metadata properties but wait until after reading
+        # metadata from a file before overriding with specified values.
+        self.id = None
+        self.name = None
+        self.origin = None
+        self.type = None
+
         if filename is not None:
             # Read the file, including metadata if present.
             self.read(filename, filetype, include_meta=True)
 
         # Set any metadata explicity specified, overriding that
         # read from the file.
-        self.id = mocid
-        self.name = name
-        self.origin = origin
-        self.type = moctype
+        if mocid is not None:
+            self.id = mocid
+        if name is not None:
+            self.name = name
+        if origin is not None:
+            self.origin = origin
+        if moctype is not None:
+            self.type = moctype
 
         # Add any cells specified in the arguments.
         if order is not None and cells is not None:
@@ -181,8 +194,8 @@ class MOC(object):
 
         >>> p = MOC(4, (5, 6))
         >>> p += MOC(4, (7, 8))
-        >>> sorted(p[4])
-        [5, 6, 7, 8]
+        >>> repr(p)
+        '<MOC: [(4, [5, 6, 7, 8])]>'
         """
 
         if not isinstance(other, MOC):
@@ -198,9 +211,8 @@ class MOC(object):
 
         Returns a MOC which is the union of two MOCs.
 
-        >>> s = MOC(4, (5, 6)) + MOC(4, (7, 8))
-        >>> sorted(s[4])
-        [5, 6, 7, 8]
+        >>> MOC(4, (5, 6)) + MOC(4, (7, 8))
+        <MOC: [(4, [5, 6, 7, 8])]>
         """
 
         if not isinstance(other, MOC):
@@ -218,11 +230,8 @@ class MOC(object):
         Returns a MOC which is the copy of the first MOC with the
         intersection with the second MOC removed.
 
-        >>> d = MOC(0, (0,)) - MOC(2, (15,))
-        >>> sorted(d[1])
-        [0, 1, 2]
-        >>> sorted(d[2])
-        [12, 13, 14]
+        >>> MOC(0, (0,)) - MOC(2, (15,))
+        <MOC: [(1, [0, 1, 2]), (2, [12, 13, 14])]>
         """
 
         if not isinstance(other, MOC):
@@ -247,6 +256,18 @@ class MOC(object):
 
         return self
 
+    def __repr__(self):
+        """Generate printable representation.
+
+        Since the constructor only accepts cells at one order and we may be
+        generating a representation for a MOC with cells at multiple orders
+        we can't try to give an expression which would construct the object.
+        Instead show a description in angle brackets.
+        """
+
+        return '<MOC: {0!r}>'.format(
+            [(o, sorted(cs)) for (o, cs) in enumerate(self._orders) if cs])
+
     @property
     def order(self):
         """The highest order at which the MOC has cells.
@@ -338,7 +359,7 @@ class MOC(object):
         58.63
         """
 
-        return self.area * ((180 / pi ) ** 2)
+        return self.area * ((180 / pi) ** 2)
 
     @property
     def cells(self):
@@ -359,7 +380,7 @@ class MOC(object):
 
         return n
 
-    def add(self, order, cells):
+    def add(self, order, cells, no_validation=False):
         """Add cells at a given order to the MOC.
 
         The cells are inserted into the MOC at the specified order.  This
@@ -375,19 +396,28 @@ class MOC(object):
         >>> m.add(5, (88, 89))
         >>> m.cells
         4
+
+        The `no_validation` option can be given to skip validation of the
+        cell numbers.  They must already be integers in the correct range.
         """
 
         self._normalized = False
 
         order = self._validate_order(order)
 
-        cell_set = set()
+        if no_validation:
+            # Simply add the given cells to the set with no validation.
+            self._orders[order].update(cells)
 
-        for cell in cells:
-            cell = self._validate_cell(order, cell)
-            cell_set.add(cell)
+        else:
+            # Collect validated cell numbers in a set for addition.
+            cell_set = set()
 
-        self._orders[order].update(cell_set)
+            for cell in cells:
+                cell = self._validate_cell(order, cell)
+                cell_set.add(cell)
+
+            self._orders[order].update(cell_set)
 
     def remove(self, order, cells):
         """Remove cells at a given order from the MOC.
@@ -421,12 +451,12 @@ class MOC(object):
 
         >>> p = MOC(4, (5, 6))
         >>> q = p.copy()
-        >>> sorted(q[4])
-        [5, 6]
+        >>> repr(q)
+        '<MOC: [(4, [5, 6])]>'
         """
 
         copy = MOC(name=self.name, mocid=self.id,
-                origin=self.origin, moctype=self.type)
+                   origin=self.origin, moctype=self.type)
 
         copy += self
 
@@ -482,9 +512,10 @@ class MOC(object):
                     # cells.  Those which actually match the area we are
                     # trying to remove will be removed at the next stage.
                     self._orders[order_i].remove(cell_i)
-                    self.add(order_i + 1, range(cell_i << 2, (cell_i + 1) << 2))
+                    self.add(order_i + 1,
+                             range(cell_i << 2, (cell_i + 1) << 2))
                 elif operation == 'inter':
-                    return (order, (cell,))
+                    return [(order, (cell,))]
 
         # Check for the specific cell itself, but only after looking at larger
         # cells because for the "remove" operation we may have broken up
@@ -495,7 +526,9 @@ class MOC(object):
             elif operation == 'remove':
                 self._orders[order].remove(cell)
             elif operation == 'inter':
-                return (order, (cell,))
+                return [(order, (cell,))]
+
+        result = []
 
         if include_smaller:
             # Check for a smaller cell (higher order) which is part
@@ -516,29 +549,28 @@ class MOC(object):
                     for cell_i in cells:
                         self._orders[order_i].remove(cell_i)
                 elif operation == 'inter':
-                    return (order_i, cells)
+                    if cells:
+                        result.append((order_i, cells))
 
         if operation == 'check':
             return False
-        else:
-            return None
+        elif operation == 'inter':
+            return result
 
     def intersection(self, other):
         """Returns a MOC representing the intersection with another MOC.
 
         >>> p = MOC(2, (3, 4, 5))
         >>> q = MOC(2, (4, 5, 6))
-        >>> i = p.intersection(q)
-        >>> sorted(i[2])
-        [4, 5]
+        >>> p.intersection(q)
+        <MOC: [(2, [4, 5])]>
         """
 
         inter = MOC()
 
         for (order, cells) in other:
             for cell in cells:
-                i = self._compare_operation(order, cell, True, 'inter')
-                if i is not None:
+                for i in self._compare_operation(order, cell, True, 'inter'):
                     inter.add(*i)
 
         return inter
@@ -606,8 +638,8 @@ class MOC(object):
                 # so that we can avoid checking the adjacent pixels.
                 if (already_contained or (order > max_order) or
                         (((pixel ^ 1) in pixels) and
-                        ((pixel ^ 2) in pixels) and
-                        ((pixel ^ 3) in pixels))):
+                         ((pixel ^ 2) in pixels) and
+                         ((pixel ^ 3) in pixels))):
 
                     pixels.discard(pixel ^ 1)
                     pixels.discard(pixel ^ 2)
@@ -704,7 +736,7 @@ class MOC(object):
             raise ValueError('Unknown MOC file type {0}'.format(filetype))
 
     def write(self, filename, filetype=None, **kwargs):
-        """Write the converage data in the MOC object to a file.
+        """Write the coverage data in the MOC object to a file.
 
         The filetype can be given or left to be inferred as for the
         read method.
@@ -780,8 +812,8 @@ class MOC(object):
             raise TypeError('MOC order must be convertable to int')
 
         if not 0 <= order <= MAX_ORDER:
-            raise ValueError('MOC order must be in range 0-{0}'.format(
-                    MAX_ORDER))
+            raise ValueError(
+                'MOC order must be in range 0-{0}'.format(MAX_ORDER))
 
         return order
 
@@ -799,7 +831,8 @@ class MOC(object):
             raise TypeError('MOC cell must be convertable to int')
 
         if not 0 <= cell < max_cells:
-            raise ValueError('MOC cell order ' +
-                '{0} must be in range 0-{1}'.format(order, max_cells - 1))
+            raise ValueError(
+                'MOC cell order {0} must be in range 0-{1}'.format(
+                    order, max_cells - 1))
 
         return cell
